﻿using UnityEngine;
using UnityEngine.Networking;
using UnityEngine.Assertions;

using System;
using System.IO;
using System.Linq;
using System.Collections;
using System.Collections.Generic;

using haxe.root;
//using HaxeSpeedTest ;

/**
 * Blender blend shape value range: 0.0-1.0
 * Unity blend shape value range: 0.0-100.0
 */
[RequireComponent(typeof(SkinnedMeshRenderer))]
public class MaryTTSController : MonoBehaviour {

	//
	// MaryTTS server info
	public enum MaryServer {
		LOCAL,
		DFKI,
		DECAD
	};
		
	private static Dictionary<MaryServer, string> server_addresses = new Dictionary<MaryServer, string>
	{
		{MaryServer.LOCAL, "localhost:59125"}, 
		{MaryServer.DFKI,  "mary.dfki.de:59125"}, 
		{MaryServer.DECAD,  "decad.sb.dfki.de:59125"},
	};

	public MaryServer server;

	//
	// MaryTTS voice info
	public enum MaryTTSVoice {
		DFKI_PRUDENCE,
		DFKI_PRUDENCE_HSMM,
		DFKI_POPPY,
		DFKI_POPPY_HSMM,
		CMU_SLT,
		CMU_SLT_HSMM
	};

	private static Dictionary<MaryTTSVoice, string> voices = new Dictionary<MaryTTSVoice, string>
	{
		{MaryTTSVoice.DFKI_PRUDENCE, "dfki-prudence"},
		{MaryTTSVoice.DFKI_PRUDENCE_HSMM, "dfki-prudence-hsmm"},
		{MaryTTSVoice.DFKI_POPPY, "dfki-poppy"},
		{MaryTTSVoice.DFKI_POPPY_HSMM, "dfki-poppy-hsmm"},
		{MaryTTSVoice.CMU_SLT,"cmu-slt"},
		{MaryTTSVoice.CMU_SLT_HSMM,"cmu-slt-hsmm"}
	} ; //= "cmu-rms-hsmm"

	public MaryTTSVoice mary_tts_voice;

	// A global multiplier to amplify, attenuate mouth articulation.
	public double blendshapesMultiplier = 1.0;
	

	// URL composition instructions at: http://mary.dfki.de:59125/documentation.html
	private const string MARY_TTS_AUDIO_PARAMETER = "&OUTPUT_TYPE=AUDIO&AUDIO=WAVE_FILE&LOCALE=en_US";
	private const string MARY_TTS_REALISED_DURATION_PARAMETER = "&OUTPUT_TYPE=REALISED_DURATIONS&LOCALE=en_US";

	private SkinnedMeshRenderer skinnedMeshRenderer;
	private Mesh skinnedMesh;

	private MaryTTSBlendSequencer sequencer = new MaryTTSBlendSequencer();
	private double[] viseme_weights = new double[MaryTTSBlendSequencer.get_viseme_count()];

	private AudioClip audioClip = null;

	void Awake () {
		skinnedMeshRenderer = GetComponent<SkinnedMeshRenderer> ();
		skinnedMesh = GetComponent<SkinnedMeshRenderer> ().sharedMesh;
	}

	void Start () {
		Assert.IsNotNull(skinnedMeshRenderer); 
		Assert.IsNotNull(skinnedMesh); 
		// TODO -- check that all the phonemes required by the Sequencer are indeed present in the mesh.
//		for (int i = 0; i < skinnedMesh.blendShapeCount; i++) {
//			string bs_name = skinnedMesh.GetBlendShapeName(i);
//		}

	}

	public void MaryTTSspeak(string text) {
		StartCoroutine (ProcessInputText (text));
	}

	private IEnumerator ProcessInputText(string text) {
		//text = "if everything goes right both individuals shake hands and go back to the world with a smile";
		Debug.Log("Your sentence: " + text);

		yield return RetrieveMaryTTSdata(text);

		// Play audio
		AudioSource.PlayClipAtPoint(audioClip, transform.position);
		this.sequencer.reset_timers();
	}

	private IEnumerator RetrieveMaryTTSdata(string text) {
		text = text.Replace(" ", "+");


		String MARY_TTS_HTTP_ADDRESS = "http://" + MaryTTSController.server_addresses[this.server];
		// Debug.Log ("Composed address: " + MARY_TTS_HTTP_ADDRESS);

		String voice_parameter = "&VOICE=" + MaryTTSController.voices[this.mary_tts_voice];
		// Debug.Log("Selected voice: " + voice_parameter);

		string request_url = MARY_TTS_HTTP_ADDRESS + "/process?INPUT_TEXT=" + text + "&INPUT_TYPE=TEXT" + voice_parameter;
		Debug.Log("Request URL: " + request_url);
			
		// Fetch audio
		WWW audioResponse = new WWW(request_url + MARY_TTS_AUDIO_PARAMETER);
		yield return MaryTTWaitForRequest(audioResponse);

		this.audioClip = audioResponse.GetAudioClip(false, false, AudioType.WAV);

		// Fetch realised durations
		WWW rdurationsResponse = new WWW(request_url + MARY_TTS_REALISED_DURATION_PARAMETER);
		yield return MaryTTWaitForRequest (rdurationsResponse);

		lock (sequencer) {
//			string[] lines = rdurationsResponse.text.Split(new string[] { Environment.NewLine }, StringSplitOptions.None);
//			for (int i = 1; i < lines.Length; i++) {
//				Debug.Log (" === " + lines [i]);
//			}
			sequencer.parse_realized_durations(rdurationsResponse.text);
		}

	}

	IEnumerator MaryTTWaitForRequest(WWW www) {
		yield return www;

		// Check for errors
		if (!string.IsNullOrEmpty(www.error)) {
            Debug.Log("WWW error: " + www.error);
		}
	}
	
	// Update is called once per frame
	void Update() {
		this.sequencer.update(Time.time, this.viseme_weights);
		// Debug.Log (this.viseme_weights [0]);


		for (int i=0 ; i < MaryTTSBlendSequencer.get_viseme_count() ; i++) {
			string viseme = (string)(MaryTTSBlendSequencer.VISEMES [i]);

			int blendShapeIdx = this.skinnedMesh.GetBlendShapeIndex(viseme);
			// Debug.Log ("Looking for viseme " + viseme+". Index: " + blendShapeIdx);



			// Simple version
			double weight = this.viseme_weights[i] * 100.0 * this.blendshapesMultiplier;

			// Tries to soften movements at low values
			//double sw = this.viseme_weights[i] ;
			//sw = -2.0 * (sw * sw * sw) + 3.0 * (sw * sw);
			//double weight = sw * 100.0 * this.blendshapesMultiplier;

			skinnedMeshRenderer.SetBlendShapeWeight(blendShapeIdx, (float) weight);
		}

	}

	//	
	public bool IsMaryTTSspeaking() {
		return this.sequencer.is_speaking();
	}

}
