﻿using System;
using UnityEngine;

namespace HaxeSpeedTest {

	public class EyeHeadGazeLogic
	{
		private PerfMeter perf_meter = new PerfMeter("EyeHeadGazeLogicCSManual", 10.0f, true);

		private const string EYES_RIGHT = "Expressions_eyesHoriz_min";
		private const string EYES_LEFT = "Expressions_eyesHoriz_max";
		private const string EYES_DOWN = "Expressions_eyesVert_min";
		private const string EYES_UP = "Expressions_eyesVert_max";

		public static Vector3 FORWARD = new Vector3(0, 0, -1) ;
		public static Vector3 HORIZ_PLANE_NORMAL = new Vector3(0, 1, 0);
		public static Vector3 RIGHT = new Vector3(-1, 0, 0) ;

		/** Set the axes reference system. They must reflect where how the character is oriented in its initial pose.
		 */
		public static void SetReferenceAxes(Vector3 fwd, Vector3 up, Vector3 right) {
			FORWARD = fwd ;
			HORIZ_PLANE_NORMAL = up ;
			RIGHT = right ;
		}


		/** Returns the name of the four blendSHapes used to drive the eyes. In order:
		 * EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP.
		 */
		public static string[] VISEMES = new string[] { EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP };

		/** Returns the number of Visemes which this module expects to handle.*/
		public static int get_viseme_count()
		{
			return VISEMES.Length;
		}


		// Determines in degree at which point the neck should rotate
		// In more extreme angles the neck will rotate otherwise only the eyes
		private const float EYES_MAX_YAW = 30.0f;
		private const float EYES_MAX_PITCH = 12.0f;

		// Neck rotation limits (degs)
		// https://www.livestrong.com/article/95456-normal-neck-range-motion/
		private const float NECK_MAX_PITCH = 45.0f;  // Neck extension (45..70)
		private const float NECK_MIN_PITCH = -40.0f;  // Neck flexion ( 40..60)
		private const float NECK_MAX_YAW = 60.0f;  // Neck rotation (60..80)
		private const float NECK_MIN_YAW = -60.0f;  // Neck rotation
		// Unused (for the moment)
		//	private const float NECK_MAX_ROLL = 45.0f;  // Neck lateral bending
		//	private const float NECK_MIN_ROLL = -45.0f;  // Neck lateral bending

		/** The speed or rotation (deg/sec) of the eyes during fixation. */
		public float eyesRotSpeed = 1000.0f;

		/** Current eyes pitch (degs, vertical). */
		private float eyesPitch = 0.0f ;
		/** Current eyes yaw (degs, horizontal). */
		private float eyesYaw = 0.0f;

		/** Whether to enable or not the neck rotation. */
		public bool enableNeckRotation = true;

		/**  The time (secs) the neck needs to re-align when the target is beyond eyes limits. */
		public float neckRotTime = 0.1f;

		/** The maximum rotation speed (deg/sec) of the neck.
		 * Here: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4616101/
		 * They consider slow/high rotation speed for the neck as 90 - 180 deg/sec
		 */
		public float neckRotMaxSpeed = 135.0f;

		/** Current neck pitch (degs). */
		private float neckYaw = 0.0f;
		/** Current neck yaw (degs). */
		private float neckPitch = 0.0f;

		// How strongly the neck tends to reset ita rotation when the target is in the rotation range of the eyes
		public float neckRestTendencyStrenght = 0.25f;

		/** The target point to fixate. */
		public Nullable<Vector3> eyeGazeTargetPoint = null;

		private Quaternion aPoseLeftEyeRotation;
		private Quaternion aPoseRightEyeRotation;
		private Quaternion aPoseNeckRotation;

		public EyeHeadGazeLogic (Quaternion aPoseLeftEyeRotation, Quaternion aPoseRightEyeRotation, Quaternion aPoseNeckRotation)
		{
			this.aPoseLeftEyeRotation = aPoseLeftEyeRotation;
			this.aPoseRightEyeRotation = aPoseRightEyeRotation;
			this.aPoseNeckRotation = aPoseNeckRotation;
		}

		/** Brings beck neck and eyes at the default position. */
		public void ResetEyeGaze() {
			this.eyesPitch = 0.0f;
			this.eyesYaw = 0.0f;
			this.neckPitch = 0.0f;
			this.neckYaw = 0.0f;
		}

		/** Starts moving eyes and head to watch a spcified position.
		 */
		public void LookAtPoint(Vector3 targetPoint){
			//Debug.Log (eyeGazeTargetPoint + "    " + currentEyeGazePoint); 
			this.eyeGazeTargetPoint = targetPoint;
		}

		//
		// Private Methods
		//

		private static float SmoothtInc(float delta, float time, float max_speed, float dt) {
			if (Math.Abs(delta) < 0.0001f) {  // Otherwise we generate numerical jumps
				return 0.0f;
			}

			float speed = delta / time;
			if (speed > max_speed) {
				speed = max_speed;
			} else if (speed < -max_speed) {
				speed = -max_speed;
			}
			float inc = speed * dt;
			//		if (Mathf.Abs (inc) > Mathf.Abs (delta)) {
			//			Debug.Log ("RotSmoothtInc: incrementing too much!!!");
			//		}
			return inc;
		}

		private static float LinearInc(float delta, float speed, float dt) {
			float inc = speed * dt;
			inc = Mathf.Min (inc, Mathf.Abs (delta));
			inc *= Mathf.Sign (delta);
			return inc;
		}


		//sets the eyes to look at the target
		// target_point is the absolute point in space that we want to wantch at.
		public void UpdateEyesRotation(Vector3 leftEyePosition, Vector3 rightEyePosition,
			Quaternion leftEyeRotation, Quaternion rightEyeRotation,
			ref float[] outBlendShapeWeights,
			ref Quaternion inOutNeckRotation,
			float dt)
		{
			perf_meter.iterStart ();

			if (this.eyeGazeTargetPoint == null) {
				return;
			}

			Vector3 abs_target_point = this.eyeGazeTargetPoint.Value; 

			//
			// Calculate the target in "eye-apose" space, i.e., relatively to the eye (or their midpoint) when the character stands in A-Pose.
			// It means that the target_point must be transformed as if it positioned relatively to the character standing in the default A-Pose.

			//
			// Decomposing the eye rotation matrix
			// The current rotation C can be decomposed in initial rotation I and an additional delta D in global space.
			//   C = D * I
			// (That is different from decomposing in initial plus local transform C = I * D_loc)
			// Hence,
			//   D = C * Iinv
			// In order to transform the absolute target in the original global space before rotation, we need the inverse of D.
			// In general (AB)inv = Binv * Ainv
			// Hence:
			//  Dinv = I * Cinv
			//


			//
			// This will be splitted in two when the two eyes will be driven independently.

			Vector3 eyeMidPoint = new Vector3(
				(leftEyePosition.x + rightEyePosition.x) / 2.0f, 
				(leftEyePosition.y + rightEyePosition.y) / 2.0f, 
				(leftEyePosition.z + rightEyePosition.z) / 2.0f
			);


			Vector3 targetPointInvDelta = this.aPoseLeftEyeRotation * Quaternion.Inverse(leftEyeRotation) * abs_target_point;

			Vector3 eyeMidPointInvDelta = this.aPoseLeftEyeRotation * Quaternion.Inverse(leftEyeRotation) * eyeMidPoint ;
			Vector3 targetPointEyeSpace = targetPointInvDelta - eyeMidPointInvDelta;


			// Calculate the targetPoint projected on the horizontal plane which passes through the eyesMidPoint
			//Vector3 targetHplaneProjection = new Vector3(targetPointEyeSpace.x, 0, targetPointEyeSpace.z) ;
			float hproj_dist = targetPointEyeSpace.x * HORIZ_PLANE_NORMAL.x
				+ targetPointEyeSpace.y * HORIZ_PLANE_NORMAL.y
				+ targetPointEyeSpace.z * HORIZ_PLANE_NORMAL.z ;
			//var targetHplaneProjection: Vec3 = targetPointEyeSpace.subtract(Vec3.multiplyScalarOp(HORIZ_PLANE_NORMAL, hproj_dist)) ;
			Vector3 targetHplaneProjection = targetPointEyeSpace - (HORIZ_PLANE_NORMAL * hproj_dist) ;



			//
			// Calculate the pitch: the vertical orientation of the eyes.

			// Calc the angle (in degrees) between the vector pointing to the target and its projection on the vertical plane
			float vertAngleBetween = Vector3.Angle(targetPointEyeSpace, targetHplaneProjection);
			//float newEyesPitch = targetPointEyeSpace.y > 0 ? vertAngleBetween : -vertAngleBetween;
			float newEyesPitch = Vector3.Dot(targetPointEyeSpace, HORIZ_PLANE_NORMAL) >= 0 ? vertAngleBetween : -vertAngleBetween;

			//
			// Calculate the yaw: the horizontal orientation of the eyes

			// Calc the angle (in degrees) between the forward vector the the target in eye-space.
			float horizAngleBetween = Vector3.Angle(FORWARD, targetHplaneProjection);
			float newEyesYaw = Vector3.Dot(targetHplaneProjection, RIGHT) <= 0 ? horizAngleBetween : -horizAngleBetween;


			// Increment the current eyes yaw/pitch according to eyes movement speed
			float deltaEyesPitch = newEyesPitch - this.eyesPitch;
			this.eyesPitch += LinearInc (deltaEyesPitch, this.eyesRotSpeed, dt);

			float deltaEyesYaw = newEyesYaw - this.eyesYaw;
			this.eyesYaw += LinearInc (deltaEyesYaw, this.eyesRotSpeed, dt);

			//
			// Activate the yaw/pitch rotations for the neck.
			if (enableNeckRotation) {

				float deltaPitch = 0.0f;
				if (eyesPitch > EYES_MAX_PITCH) {
					deltaPitch = eyesPitch - EYES_MAX_PITCH;
					eyesPitch = EYES_MAX_PITCH;
				} else if (eyesPitch < -EYES_MAX_PITCH) {
					deltaPitch = eyesPitch + EYES_MAX_PITCH;
					eyesPitch = -EYES_MAX_PITCH;
				}

				neckPitch += SmoothtInc (deltaPitch, neckRotTime, neckRotMaxSpeed, dt);
				neckPitch += SmoothtInc(-neckPitch, neckRotTime / neckRestTendencyStrenght, neckRotMaxSpeed * neckRestTendencyStrenght, dt) ;

				if (neckPitch > NECK_MAX_PITCH) {
					neckPitch = NECK_MAX_PITCH ;
				} else if(neckPitch < NECK_MIN_PITCH) {
					neckPitch = NECK_MIN_PITCH ;
				}

				// Debug.Log("Delta pitch/Neck pitch: "+deltaPitch+"\t"+neckPitch) ;


				float deltaYaw = 0.0f;
				if (eyesYaw >= EYES_MAX_YAW) {
					deltaYaw = eyesYaw - EYES_MAX_YAW;
					eyesYaw = EYES_MAX_YAW;
				} else if (eyesYaw <= -EYES_MAX_YAW) {
					deltaYaw = eyesYaw + EYES_MAX_YAW;
					eyesYaw = -EYES_MAX_YAW;
				}

				neckYaw += SmoothtInc(deltaYaw, neckRotTime, neckRotMaxSpeed, dt) ;
				neckYaw += SmoothtInc(-neckYaw, neckRotTime / neckRestTendencyStrenght, neckRotMaxSpeed * neckRestTendencyStrenght, dt) ;

				if (neckYaw > NECK_MAX_YAW) {
					neckYaw = NECK_MAX_YAW ;
				} else if(neckYaw < NECK_MIN_YAW) {
					neckYaw = NECK_MIN_YAW ;
				}

				// Debug.Log("Delta yaw/Neck yaw: "+deltaYaw+"\t"+neckYaw) ;

				// Update the local neck transform, composing its initial rotation with the current neck angles.
				//outNeckLocalRotation = this.aPoseNeckLocalRotation * Quaternion.Euler (neckYaw, -neckPitch, 0);

				Quaternion neckAbsDelta = inOutNeckRotation * Quaternion.Inverse(this.aPoseNeckRotation) ;
				
				Quaternion yawQuat = Quaternion.AngleAxis (-neckYaw, HORIZ_PLANE_NORMAL);
				Quaternion yawGlobal = neckAbsDelta * yawQuat * Quaternion.Inverse (neckAbsDelta);
				
				Quaternion pitchQuat = Quaternion.AngleAxis (-neckPitch, RIGHT);
				Quaternion pitchGlobal = neckAbsDelta * pitchQuat * Quaternion.Inverse (neckAbsDelta);

				inOutNeckRotation = pitchGlobal * yawGlobal * aPoseNeckRotation;

			}

			//
			// Calculate the proportion from the pitch angle to BlendShape percentage
			// weightEyesPitch : 50% = pitch : 45deg
			float weightEyesPitch = 0.5f * eyesPitch / 45.0f;

			// Calculate the proportion from the yaw angle to the BlendShape percentage
			// weightEyesYaw : 50% = yaw : 52deg
			float weightEyesYaw = 0.5f * eyesYaw / 52.0f;

			//Debug.Log("weightEyesDown: " + weightEyesPitch + " weightEyesLeft: " + weightEyesYaw);
			//UpdateEyeBlendShapeWeights(weightEyesPitch, weightEyesYaw);

			//Blendshapes are registered int the DefaultExecutionOrder: EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP
			if (weightEyesPitch >= 0.0f) {
				outBlendShapeWeights[3] = weightEyesPitch;
				outBlendShapeWeights[2] = 0.0f;
			} else {
				outBlendShapeWeights[2] = -weightEyesPitch;
				outBlendShapeWeights[3] = 0.0f;
			}

			if (weightEyesYaw >= 0.0f) {
				outBlendShapeWeights[0] = weightEyesYaw;
				outBlendShapeWeights[1] = 0.0f;
			} else {
				outBlendShapeWeights[1] = -weightEyesYaw;
				outBlendShapeWeights[0] = 0.0f;
			}

			perf_meter.iterEnd ();

		} // end update

	}
}
