//
// Tools to implement a temporal blinking of the eyes.

import haxe.ds.Vector;


class BlinkException {
    public var reason: String ;

    public function new(reason: String) {
        this.reason = reason ;
    }
}

enum BlinkStatus {
    OPENING; CLOSING; WAITING ;
}

/** .
 */
class EyeBlinker {
   	#if PERF_REPORT
	private var perfmeter: PerfMeter = new PerfMeter("EyeBlinker", 10) ;
	#end


    static var EYE_CLOSE_SPEED: Float = 8;  // The speed to close the eyes. In units/sec.
    static var EYE_OPEN_SPEED: Float = 10;

    // https://www.ncbi.nlm.nih.gov/pubmed/23403736
    static var MIN_DELAY: Float = 4.0 ;
    static var MAX_DELAY: Float = 8.0 ;


    static var VISEMES: Array<String> = [
          // For MBLab v1.5
          //'Expressions_eye02L_max',
          //'Expressions_eye02R_max'
          // For MBLab v1.6
          'Expressions_eyeClosedL_max',
          'Expressions_eyeClosedR_max'
            ] ;

    /** Returns the number of Visemes which this module expects to handle.*/
    static function get_viseme_count(): Int {
        return VISEMES.length;
    }

    private var last_time: Float = -1 ;

    private var current_weight: Float = 0.0 ;

    private var next_blink_time: Float = 0.0 ;
    private var blink_status: BlinkStatus = BlinkStatus.WAITING ;


    //
    // CONSTRUCTOR
    //
    public function new() {
    }



    /** The high-frequency routine to call 30+ times per second.

        @param now the current time, in seconds.
        @param viseme_weights An array of current weights of the viseme blend shapes.
         Values will be updated according to the synthesis time.
         The size of this vector must be the same of the module static vector of visemes.
         Values will be clamped in the range [0,1].
        @returns None
     */
    public function update(now: Float, viseme_weights: Vector<Float>) {

		#if PERF_REPORT
		this.perfmeter.iterStart() ;
		#end

        if (viseme_weights.length != VISEMES.length) {
            throw new BlinkException("Viseme weight vector has wrong size."
                                   + " Expected "+VISEMES.length+", found "+viseme_weights.length) ;
        }

        if (this.last_time == -1) {
            this.last_time = now ;
            return ;
        }

        var delta_time: Float = now - this.last_time ;

        this.last_time = now ;

        //
        //
        if(this.blink_status == BlinkStatus.WAITING) {
            if(now >= this.next_blink_time) {
                this.blink_status = BlinkStatus.CLOSING ;
            }
        } else if(this.blink_status == BlinkStatus.CLOSING) {
            var delta_close: Float = EYE_CLOSE_SPEED * delta_time ;
            this.current_weight += delta_close ;
            if(this.current_weight >= 1.0) {
                this.current_weight = 1.0 ;
                this.blink_status = BlinkStatus.OPENING;
            }
        } else if(this.blink_status == BlinkStatus.OPENING) {
            var delta_open: Float = EYE_OPEN_SPEED * delta_time ;
            this.current_weight -= delta_open ;
            if(this.current_weight <= 0.0) {
                this.current_weight = 0.0 ;
                this.blink_status = BlinkStatus.WAITING ;
                // Compute next closing time
                var delay: Float = MIN_DELAY + (MAX_DELAY - MIN_DELAY) * Math.random() ;
                this.next_blink_time = now + delay ;
            }
        }

        //
        // UPDATE THE WEIGHTS VECTOR
        for (i in 0...VISEMES.length) {
            // var viseme: String = VISEMES[i] ;
            // update the weight
            viseme_weights[i] = this.current_weight ;
        }

		#if PERF_REPORT
		this.perfmeter.iterEnd() ;
		#end

    }  // end update

}
