import glm.Vec3;
import glm.Quat;

import haxe.ds.Vector;

import haxe.io.Bytes;

class EyeHeadGazeLogic
{
	#if PERF_REPORT
	private var perfmeter: PerfMeter = new PerfMeter("EyeHeadGaze", 10, true) ;
	#end

	private static var EYES_RIGHT: String = "Expressions_eyesHoriz_min";
	private static var EYES_LEFT: String = "Expressions_eyesHoriz_max";
	private static var EYES_DOWN: String = "Expressions_eyesVert_min";
	private static var EYES_UP: String = "Expressions_eyesVert_max";


    public static var FORWARD: Vec3 = new Vec3(0, 0, -1) ;
	public static var HORIZ_PLANE_NORMAL: Vec3 = new Vec3(0, 1, 0);
	public static var RIGHT: Vec3 = new Vec3(-1, 0, 0) ;

	/** Set the axes reference system. They must reflect where how the character is oriented in its initial pose.
	 */
	public static function SetReferenceAxes(fwd: Vec3, up: Vec3, right: Vec3) {
		FORWARD = fwd ;
		HORIZ_PLANE_NORMAL = up ;
		RIGHT = right ;
	}


	/** Returns the name of the four blendSHapes used to drive the eyes. In order:
	 * EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP.
	 */
	// public static function GetBlendShapeNames(): Array<String> {
    //     return [EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP];
	// }

	static var VISEMES: Array<String> = [EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP];
    /** Returns the number of Visemes which this module expects to handle.*/
    static function get_viseme_count(): Int {
        return VISEMES.length;
    }



	// Determines in degree at which point the neck should rotate
	// In more extreme angles the neck will rotate otherwise only the eyes
	private static var EYES_MAX_YAW: Float = 30.0 ;
	private static var EYES_MAX_PITCH: Float = 12.0 ;

	// Neck rotation limits (degs)
	// https://www.livestrong.com/article/95456-normal-neck-range-motion/
	private static var NECK_MAX_PITCH: Float = 45.0;  // Neck extension (45..70)
	private static var NECK_MIN_PITCH: Float = -40.0;  // Neck flexion ( 40..60)
	private static var NECK_MAX_YAW: Float = 60.0;  // Neck rotation (60..80)
	private static var NECK_MIN_YAW: Float = -60.0;  // Neck rotation
	// Unused (for the moment)
	//	static var NECK_MAX_ROLL: Float = 45.0f;  // Neck lateral bending
	//	static var NECK_MIN_ROLL: Float = -45.0f;  // Neck lateral bending

	/** The speed or rotation (deg/sec) of the eyes during fixation. */
	public var eyesRotSpeed: Float = 1000.0;

	/** Current eyes pitch (degs, vertical). */
	private var eyesPitch: Float = 0.0 ;
	/** Current eyes yaw (degs, horizontal). */
	private var eyesYaw: Float = 0.0 ;

	/** Whether to enable or not the neck rotation. */
	public var enableNeckRotation: Bool = true;

	/**  The time (secs) the neck needs to re-align when the target is beyond eyes limits. */
	public var neckRotTime: Float = 0.1;

	/** The maximum rotation speed (deg/sec) of the neck.
	 * Here: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4616101/
	 * They consider slow/high rotation speed for the neck as 90 - 180 deg/sec
	 */
	public var neckRotMaxSpeed: Float = 135.0;

	/** Current neck pitch (degs). */
	private var neckYaw: Float = 0.0;
	/** Current neck yaw (degs). */
	private var neckPitch: Float = 0.0;

	// How strongly the neck tends to reset ita rotation when the target is in the rotation range of the eyes
	public var neckRestTendencyStrenght: Float = 0.25;

	/** The target point to fixate. */
	private var eyeGazeTargetPoint: Vec3 = null;

	private var aPoseLeftEyeRotation: Quat;
	private var aPoseRightEyeRotation: Quat;
	private var aPoseNeckRotation: Quat;


	public function new(aPoseLeftEyeRotation: Quat, aPoseRightEyeRotation: Quat, aPoseNeckRotation: Quat)
	{
		this.aPoseLeftEyeRotation = aPoseLeftEyeRotation;
		this.aPoseRightEyeRotation = aPoseRightEyeRotation;
		this.aPoseNeckRotation = aPoseNeckRotation;
	}


	/** Brings beck neck and eyes at the default position. */
	public function ResetEyeGaze(): Void {
		this.eyesPitch = 0.0;
		this.eyesYaw = 0.0;
		this.neckPitch = 0.0;
		this.neckYaw = 0.0;
	}

	/** Starts moving eyes and head to watch a spcified position.
	 */
	public function LookAtPoint(targetPoint: Vec3): Void{
		//Debug.Log (eyeGazeTargetPoint + "    " + currentEyeGazePoint); 
		this.eyeGazeTargetPoint = targetPoint;
	}

	//
	// Private Methods
	//

	private static function SmoothtInc(delta: Float, time: Float, max_speed: Float, dt: Float): Float {
		if (Math.abs(delta) < 0.0001) {  // Otherwise we generate numerical jumps
			return 0.0;
		}

		var speed:Float = delta / time;
		if (speed > max_speed) {
			speed = max_speed;
		} else if (speed < -max_speed) {
			speed = -max_speed;
		}
		var inc: Float = speed * dt;
		//		if (Mathf.Abs (inc) > Mathf.Abs (delta)) {
		//			Debug.Log ("RotSmoothtInc: incrementing too much!!!");
		//		}
		return inc;
	}

	private static function LinearInc(delta: Float, speed: Float, dt: Float): Float {
		var inc: Float = speed * dt;
		inc = Math.min (inc, Math.abs (delta));
		//inc *= Math.Sign (delta);
        if (delta < 0) {
            inc *= -1 ;
        }
		// trace("LinerInc: delta: "+delta+"\tspeed: "+speed+"\tdt: "+dt+"\tinc: "+inc);
		return inc;
	}


	//sets the eyes to look at the target
	// target_point is the absolute point in space that we want to wantch at.
	public function UpdateEyesRotation(leftEyePosition: Vec3, rightEyePosition: Vec3,
		leftEyeRotation: Quat, rightEyeRotation: Quat,
		outBlendShapeWeights: Vector<Float>,
		inOutNeckRotation: Quat,
		dt: Float): Void
    {
		#if PERF_REPORT
		this.perfmeter.iterStart() ;
		#end


		if (this.eyeGazeTargetPoint == null) {
			return;
		}

		var abs_target_point: Vec3 = this.eyeGazeTargetPoint; 

        //
		// Calculate the target in "eye-apose" space, i.e., relatively to the eye (or their midpoint) when the character stands in A-Pose.
		// It means that the target_point must be transformed as if it positioned relatively to the character standing in the default A-Pose.

		//
		// Decomposing the eye rotation matrix
		// The current rotation C can be decomposed in initial rotation I and an additional delta D in global space.
		//   C = D * I
		// (That is different from decomposing in initial plus local transform C = I * D_loc)
		// Hence,
		//   D = C * Iinv
		// In order to transform the absolute target in the original global space before rotation, we need the inverse of D.
		// In general (AB)inv = Binv * Ainv
		// Hence:
		//  Dinv = I * Cinv
		//

		//
		// This will be splitted in two when the two eyes will be driven independently.

		var eyeMidPoint: Vec3 = new Vec3(
			(leftEyePosition.x + rightEyePosition.x) / 2.0, 
			(leftEyePosition.y + rightEyePosition.y) / 2.0, 
			(leftEyePosition.z + rightEyePosition.z) / 2.0
		);

   		//Vector3 targetPointInvDelta = this.aPoseLeftEyeRotation * Quaternion.Inverse(leftEyeRotation) * abs_target_point;
		var targetPointInvDelta: Vec3 = this.aPoseLeftEyeRotation.rotate(leftEyeRotation.inverted().rotate(abs_target_point));

		// Vector3 eyeMidPointInvDelta = this.aPoseLeftEyeRotation * Quaternion.Inverse(leftEyeRotation) * eyeMidPoint ;
		var eyeMidPointInvDelta: Vec3 = this.aPoseLeftEyeRotation.rotate(leftEyeRotation.inverted().rotate(eyeMidPoint)) ;
		// Vector3 targetPointEyeSpace = targetPointInvDelta - eyeMidPointInvDelta;
		var targetPointEyeSpace: Vec3 = targetPointInvDelta.subtract(eyeMidPointInvDelta);
		// trace("targetPointEyeSpace: "+targetPointEyeSpace);

		//
		// Calculate the targetPoint projected on the horizontal plane which passes through the eyesMidPoint

		// https://stackoverflow.com/questions/9605556/how-to-project-a-point-onto-a-plane-in-3d
		// 1) Make a vector from your orig point to the point of interest:
		// v = point-orig (in each dimension);
		// 2) Take the dot product of that vector with the unit normal vector n:
		// dist = vx*nx + vy*ny + vz*nz; dist = scalar distance from point to plane along the normal
		// 3) Multiply the unit normal vector by the distance, and subtract that vector from your point.
		// projected_point = point - dist*normal;

		// 1) In my case orig is 0 --> v==orig
		// 2)
		var hproj_dist = targetPointEyeSpace.x * HORIZ_PLANE_NORMAL.x
						+ targetPointEyeSpace.y * HORIZ_PLANE_NORMAL.y
						+ targetPointEyeSpace.z * HORIZ_PLANE_NORMAL.z ;
		// 3)
		var targetHplaneProjection: Vec3 = targetPointEyeSpace.subtract(Vec3.multiplyScalarOp(HORIZ_PLANE_NORMAL, hproj_dist)) ;

		// trace("targetHplaneProjection: "+targetHplaneProjection);

		//
		// Calculate the pitch: the vertical orientation of the eyes.
		// Calc the angle (in degrees) between the vector pointing to the target and its projection on the vertical plane
		var vertAngleBetween: Float = Vec3.angleBetween(targetPointEyeSpace, targetHplaneProjection)  * 180 / Math.PI;
		// Check if the angle is positive or negative (up/down) using the dot product with the vertical vector.
		var newEyesPitch: Float = Vec3.dot(targetPointEyeSpace, HORIZ_PLANE_NORMAL) >= 0 ? vertAngleBetween : -vertAngleBetween;


		// trace("newEyesPitch: "+newEyesPitch);

        //
		// Calculate the yaw: the horizontal orientation of the eyes
		// Calc the angle (in degrees) between the forward vector the the target in eye-space.
        var horizAngleBetween: Float = Vec3.angleBetween(FORWARD, targetHplaneProjection) * 180 / Math.PI;
		// Check if the angle is positive or negative (left/right) using the dot product with the side vector.
		var newEyesYaw: Float = Vec3.dot(targetHplaneProjection, RIGHT) <= 0 ? horizAngleBetween : -horizAngleBetween;

		// trace("newEyesYaw: "+newEyesYaw);

        //
		// Increment the current eyes yaw/pitch according to eyes movement speed
		var deltaEyesPitch: Float = newEyesPitch - this.eyesPitch;
		this.eyesPitch += LinearInc (deltaEyesPitch, this.eyesRotSpeed, dt);

		var deltaEyesYaw: Float = newEyesYaw - this.eyesYaw;
		this.eyesYaw += LinearInc (deltaEyesYaw, this.eyesRotSpeed, dt);

		// trace("Eyes pitch/yaw (degs): "+this.eyesPitch+"\t"+this.eyesYaw);

        //
		// Activate the yaw/pitch rotations for the neck.
		if (enableNeckRotation) {

			var deltaPitch: Float = 0.0;
			if (eyesPitch > EYES_MAX_PITCH) {
				deltaPitch = eyesPitch - EYES_MAX_PITCH;
				eyesPitch = EYES_MAX_PITCH;
			} else if (eyesPitch < -EYES_MAX_PITCH) {
				deltaPitch = eyesPitch + EYES_MAX_PITCH;
				eyesPitch = -EYES_MAX_PITCH;
			}

			neckPitch += SmoothtInc (deltaPitch, neckRotTime, neckRotMaxSpeed, dt);
			neckPitch += SmoothtInc(-neckPitch, neckRotTime / neckRestTendencyStrenght, neckRotMaxSpeed * neckRestTendencyStrenght, dt) ;

			if (neckPitch > NECK_MAX_PITCH) {
				neckPitch = NECK_MAX_PITCH ;
			} else if(neckPitch < NECK_MIN_PITCH) {
				neckPitch = NECK_MIN_PITCH ;
			}

			// trace("Delta pitch/Neck pitch: "+deltaPitch+"\t"+neckPitch) ;


			var deltaYaw: Float = 0.0;
			if (eyesYaw >= EYES_MAX_YAW) {
				deltaYaw = eyesYaw - EYES_MAX_YAW;
				eyesYaw = EYES_MAX_YAW;
			} else if (eyesYaw <= -EYES_MAX_YAW) {
				deltaYaw = eyesYaw + EYES_MAX_YAW;
				eyesYaw = -EYES_MAX_YAW;
			}

			neckYaw += SmoothtInc(deltaYaw, neckRotTime, neckRotMaxSpeed, dt) ;
			neckYaw += SmoothtInc(-neckYaw, neckRotTime / neckRestTendencyStrenght, neckRotMaxSpeed * neckRestTendencyStrenght, dt) ;

			if (neckYaw > NECK_MAX_YAW) {
				neckYaw = NECK_MAX_YAW ;
			} else if(neckYaw < NECK_MIN_YAW) {
				neckYaw = NECK_MIN_YAW ;
			}

			// trace("Delta yaw/Neck yaw: "+deltaYaw+"\t"+neckYaw) ;

            var neckPitchRad: Float = neckPitch * Math.PI / 180 ;
            var neckYawRad: Float = neckYaw * Math.PI / 180 ;

			// trace("Neck Pitch/Yaw degs: "+neckPitch+"\t "+neckYaw) ;
			// trace("Neck Pitch/Yaw rad: "+neckPitchRad+"\t "+neckYawRad) ;
			//Sys.stdout().write(Bytes.ofString("Neck Pitch/Yaw rad: "+neckPitchRad+"\t "+neckYawRad)) ;

			// 
			// Update the global neck transform (G), composing its initial rotation (I) with the rotation (R) composing neck angles yaw and pitch (R = Ry * Rp).
			// G = R * I

			// First compute the absolute-frame delta (D) for the initial pose to the current (C)
			// C = D * I
			// D = C * Iinv
			var neckAbsDelta: Quat = inOutNeckRotation.multiplied(this.aPoseNeckRotation.inverted()) ;

			// The rotation of the neck around the global vertical must be calculated
			// be first bringing the neck to its original position and then back to the current pose.
			// Ry = D * Y * Dinv
			var yawQuat = new Quat();
			Quat.axisAngle(HORIZ_PLANE_NORMAL, -neckYawRad, yawQuat) ;
			var yawGlobal:Quat = neckAbsDelta.multiplied(yawQuat.multiplied(neckAbsDelta.inverted())) ;

			// Rp = D * P * Dinv
			var pitchQuat = new Quat() ;
			Quat.axisAngle(RIGHT, -neckPitchRad, pitchQuat) ;
			var pitchGlobal: Quat = neckAbsDelta.multiplied(pitchQuat.multiplied(neckAbsDelta.inverted())) ;

			// G = Rp * Ry * I
			inOutNeckRotation.fromQuat(pitchGlobal.multiplied(yawGlobal.multiplied(this.aPoseNeckRotation))) ;

			// Test: set simply the initial rotation.
			//outNeckLocalRotation.fromQuat(this.aPoseNeckRotation) ;

		}



        //
		// Calculate the proportion from the pitch angle to BlendShape percentage
		// weightEyesPitch : 50% = pitch : 45deg
		var weightEyesPitch: Float = 0.5 * eyesPitch / 45.0;

		// Calculate the proportion from the yaw angle to the BlendShape percentage
		// weightEyesYaw : 50% = yaw : 52deg
		var weightEyesYaw: Float = 0.5 * eyesYaw / 52.0;

		//Blendshapes are registered int the DefaultExecutionOrder: EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP
		if (weightEyesPitch >= 0.0) {
			outBlendShapeWeights[3] = weightEyesPitch;
			outBlendShapeWeights[2] = 0.0;
		} else {
			outBlendShapeWeights[2] = -weightEyesPitch;
			outBlendShapeWeights[3] = 0.0;
		}

		if (weightEyesYaw >= 0.0) {
			outBlendShapeWeights[0] = weightEyesYaw;
			outBlendShapeWeights[1] = 0.0;
		} else {
			outBlendShapeWeights[1] = -weightEyesYaw;
			outBlendShapeWeights[0] = 0.0;
		}

		#if PERF_REPORT
		this.perfmeter.iterEnd() ;
		#end

    }

}
