import bge

from mathutils import Vector
from mathutils import Quaternion

import math

import time

USE_HAND_WRITTEN_VERSION = True

if USE_HAND_WRITTEN_VERSION:
    from eye_head_gaze import EyeHeadGazeLogic
else:
    from haxe_lib import EyeHeadGazeLogic
    from haxe_lib import glm_Vec3
    from haxe_lib import glm_Quat

#
#  GLOBAL VARIABLES

NECK_BONE = "neck"
EYE_BONE_R = "eye_R"
EYE_BONE_L = "eye_L"


#
# INITIALIZE
#
# print("CWD: '{}'. Scene directory: '{}'. ".format(os.getcwd(), scene_dir))

co = bge.logic.getCurrentController()
obj = co.owner
scene = obj.scene
obj = scene.objects["Anna"]  # Take the reference to the armature

gaze_target_obj = scene.objects["GazeTarget"]

# Set the normal to the Horizontal Plane and the forward vector for Blender.
if USE_HAND_WRITTEN_VERSION:
    EyeHeadGazeLogic.SetReferenceAxes(fwd=Vector((0, -1, 0)), up=Vector((0, 0, 1)), right=Vector((-1, 0, 0)))
else:
    EyeHeadGazeLogic.SetReferenceAxes(fwd=glm_Vec3(0, -1, 0), up=glm_Vec3(0, 0, 1), right=glm_Vec3(-1, 0, 0))

    #
    # UTILITY functions to convert from/to mathutils/glm types

    def quat_to_str(q: glm_Quat) -> str:
        return "Quat(xyzw)<{:06f} {:06f} {:06f} {:06f}>".format(q.x, q.y, q.z, q.w)


    def toQuat(q: Quaternion) -> glm_Quat:
        return glm_Quat(x=q.x, y=q.y, z=q.z, w=q.w)


    def toQuaternion(q: glm_Quat) -> Quaternion:
        return Quaternion((q.w, q.x, q.y, q.z))


    def toVec3(v: Vector) -> glm_Vec3:
        return glm_Vec3(v.x, v.y, v.z)


if USE_HAND_WRITTEN_VERSION:
    logic = EyeHeadGazeLogic(
        aPoseLeftEyeRotation=obj.channels[EYE_BONE_L].pose_matrix.to_quaternion(),
        aPoseRightEyeRotation=obj.channels[EYE_BONE_R].pose_matrix.to_quaternion(),
        aPoseNeckRotation=obj.channels[NECK_BONE].pose_matrix.to_quaternion()
    )
else:
    logic = EyeHeadGazeLogic(
        aPoseLeftEyeRotation=toQuat(obj.channels[EYE_BONE_L].pose_matrix.to_quaternion()),
        aPoseRightEyeRotation=toQuat(obj.channels[EYE_BONE_R].pose_matrix.to_quaternion()),
        aPoseNeckRotation=toQuat(obj.channels[NECK_BONE].pose_matrix.to_quaternion())
    )

logic.enableNeckRotation = True

#print("Neck aPoserot:" + quat_to_str(toQuat(obj.channels[NECK_BONE].pose_matrix.to_quaternion())))
#print("Neck aPoserot:" + str(obj.channels[NECK_BONE].pose_matrix.to_quaternion()))


aPoseNeckAbsRot = obj.channels[NECK_BONE].pose_matrix.to_quaternion()

viseme_weights = [0.0] * EyeHeadGazeLogic.get_viseme_count()  # The always up-to-date vector of weights

last_update = 0


#
# This function updates the eye-gaze of the character
def update():
    global viseme_weights
    global last_update

    if last_update == 0:
        last_update = time.time()
        return

    now = time.time()  # current time
    dt = now - last_update

    #
    # Move Eye gaze target
    gaze_position = Vector((math.sin(now * 2) * 1, -1, 1.5 + math.sin(now * 3) * 0.5))
    # print("Looking at " + str(gaze_position))
    gaze_target_obj.worldPosition = gaze_position

    #
    # Update the vector of weights and the neck rotation

    neck_rot = obj.channels[NECK_BONE].pose_matrix.to_quaternion()

    #
    # Apply
    if not USE_HAND_WRITTEN_VERSION:
        neck_rot = toQuat(neck_rot)
        gaze_position = toVec3(gaze_position)

    logic.LookAtPoint(gaze_position)

    if USE_HAND_WRITTEN_VERSION:
        logic.UpdateEyesRotation(leftEyePosition=Vector(obj.channels[EYE_BONE_L].pose_head),
                                 rightEyePosition=Vector(obj.channels[EYE_BONE_R].pose_head),
                                 leftEyeRotation=obj.channels[EYE_BONE_L].pose_matrix.to_quaternion(),
                                 rightEyeRotation=obj.channels[EYE_BONE_R].pose_matrix.to_quaternion(),
                                 outBlendShapeWeights=viseme_weights,
                                 inOutNeckRotation=neck_rot,
                                 dt=dt)
    else:

        logic.UpdateEyesRotation(leftEyePosition=toVec3(obj.channels[EYE_BONE_L].pose_head),
                                 rightEyePosition=toVec3(obj.channels[EYE_BONE_R].pose_head),
                                 leftEyeRotation=toQuat(obj.channels[EYE_BONE_L].pose_matrix.to_quaternion()),
                                 rightEyeRotation=toQuat(obj.channels[EYE_BONE_R].pose_matrix.to_quaternion()),
                                 outBlendShapeWeights=viseme_weights,
                                 inOutNeckRotation=neck_rot,
                                 dt=dt)

    # print(viseme_weights)
    # print("Neck after: " + quat_to_str(neck_rot))

    #
    # Update the position of the driving bones
    for viseme, weight in zip(logic.VISEMES, viseme_weights):
        ch = obj.channels['shapekey-'+viseme]
        ch.location = 0, weight, 0

    if not USE_HAND_WRITTEN_VERSION:
        neck_rot = toQuaternion(neck_rot)

    #
    # Updates the rotation of the neck
    # We get back the pose_space rotation, so we have to convert it in local_space
    #
    # Rpose = Rparent * Rlocal
    # Rlocal = Rparent.inv() * Rpose
    #obj.channels[NECK_BONE].rotation_quaternion = \
    #    obj.channels[NECK_BONE].parent.pose_matrix.to_quaternion().inverted() * toQuaternion(neck_local_rot)

    # print("Parent rot: " +str(obj.channels[NECK_BONE].parent.pose_matrix.to_quaternion()))

    #small_vert_rot = mathutils.Quaternion((0.0, 0.0, 1.0), math.radians(1.0))
    #obj.channels[NECK_BONE].rotation_quaternion = mathutils.Quaternion(obj.channels[NECK_BONE].rotation_quaternion) * small_vert_rot
    #obj.channels[NECK_BONE].rotation_quaternion = small_vert_rot * mathutils.Quaternion(obj.channels[NECK_BONE].rotation_quaternion)
    #obj.channels[NECK_BONE].rotation_quaternion = 1,0,0,0


    #obj.channels[NECK_BONE].rotation_quaternion = obj.channels[NECK_BONE].parent.pose_matrix.inverted().to_quaternion() * aPoseNeckAbsRot
    #obj.channels[NECK_BONE].rotation_quaternion = obj.channels[NECK_BONE].pose_matrix.to_quaternion().inverted() * aPoseNeckAbsRot

    # Inverse rotation

    # G = A * L
    # L = Ainv * G

    # L = Ginv * A
    # obj.channels[NECK_BONE].rotation_quaternion = (aPoseNeckAbsRot.inverted() * toQuaternion(neck_local_rot)).inverted()
    obj.channels[NECK_BONE].rotation_quaternion = (neck_rot.inverted() * aPoseNeckAbsRot)

    # Update the armature to move the bones (and consequently the shapekeys)
    obj.update()

    last_update = now
