
from enum import Enum
from random import random

from typing import List

from perf_utils import PerfMeter


class BlinkException(Exception):
    pass


class BlinkStatus(Enum):
    OPENING = 1
    CLOSING = 2
    WAITING = 3


class EyeBlinker:

    # The speed to close the eyes. In units/sec.
    EYE_CLOSE_SPEED = 8  # type: float
    EYE_OPEN_SPEED = 10  # type: float

    # https://www.ncbi.nlm.nih.gov/pubmed/23403736
    MIN_DELAY = 4.0  # type: float
    MAX_DELAY = 8.0  # type: float

    VISEMES = ['Expressions_eyeClosedL_max',
               'Expressions_eyeClosedR_max']


    @staticmethod
    def get_viseme_count() -> int:
        """Returns the number of Visemes which this module expects to handle."""
        return len(EyeBlinker.VISEMES)

    def __init__(self):
        self.last_time = -1  # type: float
        self.current_weight = 0.0  # type: float
        self.next_blink_time = 0.0  # type: float
        self.blink_status = BlinkStatus.WAITING  # type: BlinkStatus

        self.perf_meter = PerfMeter(prefix="BlinkerPythonManual", updateInterval=10.0, writeToFile=True)

    def update(self, now: float, viseme_weights: List[float]) -> None:
        """The high-frequency routine to call 30+ times per second.

            @param now the current time, in seconds.
            @param viseme_weights An array of current weights of the viseme blend shapes.
             Values will be updated according to the synthesis time.
             The size of this vector must be the same of the module static vector of visemes.
             Values will be clamped in the range [0,1].
            @returns None
        """

        self.perf_meter.iterStart()

        if len(viseme_weights) != len(EyeBlinker.VISEMES):
            raise BlinkException("Viseme weight vector has wrong size."
                                 + " Expected "+str(len(EyeBlinker.VISEMES))+", found "+str(len(viseme_weights)))

        if self.last_time == -1:
            self.last_time = now
            return

        delta_time = now - self.last_time  # type: float

        self.last_time = now

        if self.blink_status == BlinkStatus.WAITING:
            if now >= self.next_blink_time:
                self.blink_status = BlinkStatus.CLOSING
        elif self.blink_status == BlinkStatus.CLOSING:
            delta_close = self.EYE_CLOSE_SPEED * delta_time
            self.current_weight += delta_close
            if self.current_weight >= 1.0:
                self.current_weight = 1.0
                self.blink_status = BlinkStatus.OPENING
        elif self.blink_status == BlinkStatus.OPENING:
            delta_open = self.EYE_OPEN_SPEED * delta_time
            self.current_weight -= delta_open
            if self.current_weight <= 0.0:
                self.current_weight = 0.0
                self.blink_status = BlinkStatus.WAITING
                # Compute next closing time
                delay = self.MIN_DELAY + (self.MAX_DELAY - self.MIN_DELAY) * random()
                self.next_blink_time = now + delay

        #
        # UPDATE THE WEIGHTS VECTOR
        for i in range(len(EyeBlinker.VISEMES)):
            viseme = EyeBlinker.VISEMES[i]
            # update the weight
            viseme_weights[i] = self.current_weight

        self.perf_meter.iterEnd()
