from perf_utils import PerfMeter

from mathutils import Vector
from mathutils import Quaternion

import math

from typing import List

class EyeHeadGazeLogic:

    EYES_RIGHT = "Expressions_eyesHoriz_min"
    EYES_LEFT = "Expressions_eyesHoriz_max"
    EYES_DOWN = "Expressions_eyesVert_min"
    EYES_UP = "Expressions_eyesVert_max"


    FORWARD = None
    HORIZ_PLANE_NORMAL = None
    RIGHT = None

    @classmethod
    def SetReferenceAxes(cls, fwd: Vector, up: Vector, right: Vector) -> None:
        """Set the axes reference system. They must reflect where how the character is oriented in its initial pose."""
        cls.FORWARD = fwd
        cls.HORIZ_PLANE_NORMAL = up
        cls.RIGHT = right

    #
    # Returns the name of the four blendSHapes used to drive the eyes. In order:
    # EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP.
    VISEMES = [EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP]

    @classmethod
    def get_viseme_count(cls):
        """Returns the number of Visemes which this module expects to handle."""
        return len(cls.VISEMES)

    #
    # Determines in degree at which point the neck should rotate
    # In more extreme angles the neck will rotate otherwise only the eyes
    EYES_MAX_YAW = 30.0
    EYES_MAX_PITCH = 12.0

    # Neck rotation limits (degs)
    # https://www.livestrong.com/article/95456-normal-neck-range-motion/
    NECK_MAX_PITCH = 45.0
    NECK_MIN_PITCH = -40.0
    NECK_MAX_YAW = 60.0
    NECK_MIN_YAW = -60.0

    # The speed or rotation (deg/sec) of the eyes during fixation.
    eyesRotSpeed = 1000.0

    # Current eyes pitch (degs, vertical).
    eyesPitch = 0.0
    # Current eyes yaw (degs, horizontal).
    eyesYaw = 0.0

    # Whether to enable or not the neck rotation.
    enableNeckRotation = True

    # The time (secs) the neck needs to re-align when the target is beyond eyes limits.
    neckRotTime = 0.1

    # The maximum rotation speed (deg/sec) of the neck.
    # Here: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4616101/
    # They consider slow/high rotation speed for the neck as 90 - 180 deg/sec
    neckRotMaxSpeed = 135.0

    # Current neck pitch (degs).
    neckYaw = 0.0
    # Current neck yaw (degs).
    neckPitch = 0.0

    # How strongly the neck tends to reset ita rotation when the target is in the rotation range of the eyes
    neckRestTendencyStrenght = 0.25

    # The target point to fixate.
    eyeGazeTargetPoint = None

    #
    #
    def __init__(self, aPoseLeftEyeRotation: Quaternion, aPoseRightEyeRotation: Quaternion, aPoseNeckRotation: Quaternion):
        self.perf_meter = PerfMeter("EyeGazePythonManual", updateInterval=10.0, writeToFile=True)

        self.aPoseLeftEyeRotation = aPoseLeftEyeRotation
        self.aPoseRightEyeRotation = aPoseRightEyeRotation
        self.aPoseNeckRotation = aPoseNeckRotation

    # Brings beck neck and eyes at the default position.
    def ResetEyeGaze(self) -> None:
        self.eyesPitch = 0.0
        self.eyesYaw = 0.0
        self.neckPitch = 0.0
        self.neckYaw = 0.0


    # Starts moving eyes and head to watch a spcified position.
    def LookAtPoint(self, targetPoint: Vector) -> None:
        self.eyeGazeTargetPoint = targetPoint

    #
    # Private Methods
    #

    @staticmethod
    def SmoothtInc(delta: float, time: float, max_speed: float, dt: float) -> float:
        if abs(delta) < 0.0001:  # Otherwise we generate numerical jumps
            return 0.0

        speed = delta / time
        if speed > max_speed:
            speed = max_speed
        elif speed < -max_speed:
            speed = -max_speed

        inc = speed * dt
        # if (Mathf.Abs (inc) > Mathf.Abs (delta)) {
        #   Debug.Log ("RotSmoothtInc: incrementing too much!!!");
        # }
        return inc

    @staticmethod
    def LinearInc(delta: float, speed: float, dt: float) -> float:
        inc = speed * dt
        inc = min(inc, abs(delta))

        if delta < 0:
            inc *= -1
        # trace("LinerInc: delta: "+delta+"\tspeed: "+speed+"\tdt: "+dt+"\tinc: "+inc);
        return inc

    #
    # sets the eyes to look at the target
    # target_point is the absolute point in space that we want to wantch at.

    def UpdateEyesRotation(self, leftEyePosition: Vector, rightEyePosition: Vector,
        leftEyeRotation: Quaternion, rightEyeRotation: Quaternion,
        outBlendShapeWeights: List[float],
        inOutNeckRotation: Quaternion,
        dt: float) -> None:

        if self.eyeGazeTargetPoint is None:
            return

        self.perf_meter.iterStart()

        abs_target_point = self.eyeGazeTargetPoint

        eyeMidPoint = Vector((
            (leftEyePosition.x + rightEyePosition.x) / 2.0,
            (leftEyePosition.y + rightEyePosition.y) / 2.0,
            (leftEyePosition.z + rightEyePosition.z) / 2.0
        ))

        # var targetPointInvDelta: Vec3 = this.aPoseLeftEyeRotation.rotate(leftEyeRotation.inverted().rotate(abs_target_point));
        targetPointInvDelta = self.aPoseLeftEyeRotation * leftEyeRotation.inverted() * abs_target_point  # type: Vector

        # var eyeMidPointInvDelta: Vec3 = this.aPoseLeftEyeRotation.rotate(leftEyeRotation.inverted().rotate(eyeMidPoint)) ;
        eyeMidPointInvDelta = self.aPoseLeftEyeRotation * leftEyeRotation.inverted() * eyeMidPoint  # type: Vector

        # var targetPointEyeSpace: Vec3 = targetPointInvDelta.subtract(eyeMidPointInvDelta);
        targetPointEyeSpace = targetPointInvDelta - eyeMidPointInvDelta  # type: Vector

        #
        #
        # Calculate the targetPoint projected on the horizontal plane which passes through the eyesMidPoint

        # https://stackoverflow.com/questions/9605556/how-to-project-a-point-onto-a-plane-in-3d
        # 1) Make a vector from your orig point to the point of interest:
        # v = point-orig (in each dimension);
        # 2) Take the dot product of that vector with the unit normal vector n:
        # dist = vx*nx + vy*ny + vz*nz; dist = scalar distance from point to plane along the normal
        # 3) Multiply the unit normal vector by the distance, and subtract that vector from your point.
        # projected_point = point - dist*normal;

        # 1) In my case orig is 0 --> v==orig
        # 2)
        hproj_dist = targetPointEyeSpace.x * EyeHeadGazeLogic.HORIZ_PLANE_NORMAL.x\
                        + targetPointEyeSpace.y * EyeHeadGazeLogic.HORIZ_PLANE_NORMAL.y\
                        + targetPointEyeSpace.z * EyeHeadGazeLogic.HORIZ_PLANE_NORMAL.z
        # 3)
        # var targetHplaneProjection: Vec3 = targetPointEyeSpace.subtract(Vec3.multiplyScalarOp(HORIZ_PLANE_NORMAL, hproj_dist)) ;
        targetHplaneProjection = targetPointEyeSpace - (EyeHeadGazeLogic.HORIZ_PLANE_NORMAL * hproj_dist)


        #
        # Calculate the pitch: the vertical orientation of the eyes.
        # Calc the angle (in degrees) between the vector pointing to the target and its projection on the vertical plane
        # var vertAngleBetween: Float = Vec3.angleBetween(targetPointEyeSpace, targetHplaneProjection)  * 180 / Math.PI;
        vertAngleBetween = targetPointEyeSpace.angle(targetHplaneProjection) * 180 / math.pi
        # Check if the angle is positive or negative (up/down) using the dot product with the vertical vector.
        # var newEyesPitch: Float = Vec3.dot(targetPointEyeSpace, HORIZ_PLANE_NORMAL) >= 0 ? vertAngleBetween : -vertAngleBetween;
        newEyesPitch = vertAngleBetween if targetPointEyeSpace.dot(EyeHeadGazeLogic.HORIZ_PLANE_NORMAL) >= 0 else -vertAngleBetween

        #
        # Calculate the yaw: the horizontal orientation of the eyes
        # Calc the angle (in degrees) between the forward vector the the target in eye-space.
        # var horizAngleBetween: Float = Vec3.angleBetween(FORWARD, targetHplaneProjection) * 180 / Math.PI;
        horizAngleBetween = EyeHeadGazeLogic.FORWARD.angle(targetHplaneProjection) * 180 / math.pi
        # Check if the angle is positive or negative (left/right) using the dot product with the side vector.
        # var newEyesYaw: Float = Vec3.dot(targetHplaneProjection, RIGHT) <= 0 ? horizAngleBetween : -horizAngleBetween;
        newEyesYaw = horizAngleBetween if targetHplaneProjection.dot(EyeHeadGazeLogic.RIGHT) <= 0 else -horizAngleBetween;

        #
        # Increment the current eyes yaw/pitch according to eyes movement speed
        deltaEyesPitch = newEyesPitch - self.eyesPitch
        self.eyesPitch += EyeHeadGazeLogic.LinearInc(deltaEyesPitch, self.eyesRotSpeed, dt)

        deltaEyesYaw = newEyesYaw - self.eyesYaw
        self.eyesYaw += EyeHeadGazeLogic.LinearInc(deltaEyesYaw, self.eyesRotSpeed, dt)

        #
        # Activate the yaw/pitch rotations for the neck.
        if self.enableNeckRotation:
            deltaPitch = 0.0
            if self.eyesPitch > EyeHeadGazeLogic.EYES_MAX_PITCH:
                deltaPitch = self.eyesPitch - EyeHeadGazeLogic.EYES_MAX_PITCH
                self.eyesPitch = EyeHeadGazeLogic.EYES_MAX_PITCH
            elif self.eyesPitch < - EyeHeadGazeLogic.EYES_MAX_PITCH:
                deltaPitch = self.eyesPitch + EyeHeadGazeLogic.EYES_MAX_PITCH
                self.eyesPitch = - EyeHeadGazeLogic.EYES_MAX_PITCH

            self.neckPitch += EyeHeadGazeLogic.SmoothtInc(deltaPitch, self.neckRotTime, self.neckRotMaxSpeed, dt)
            self.neckPitch += EyeHeadGazeLogic.SmoothtInc(-self.neckPitch, self.neckRotTime / self.neckRestTendencyStrenght, self.neckRotMaxSpeed * self.neckRestTendencyStrenght, dt)

            if self.neckPitch > EyeHeadGazeLogic.NECK_MAX_PITCH:
                self.neckPitch = EyeHeadGazeLogic.NECK_MAX_PITCH
            elif self.neckPitch < EyeHeadGazeLogic.NECK_MIN_PITCH:
                self.neckPitch = EyeHeadGazeLogic.NECK_MIN_PITCH

            deltaYaw = 0.0
            if self.eyesYaw >= EyeHeadGazeLogic.EYES_MAX_YAW:
                deltaYaw = self.eyesYaw - EyeHeadGazeLogic.EYES_MAX_YAW
                self.eyesYaw = EyeHeadGazeLogic.EYES_MAX_YAW
            elif self.eyesYaw <= - EyeHeadGazeLogic.EYES_MAX_YAW:
                deltaYaw = self.eyesYaw + EyeHeadGazeLogic.EYES_MAX_YAW
                self.eyesYaw = - EyeHeadGazeLogic.EYES_MAX_YAW

            self.neckYaw += EyeHeadGazeLogic.SmoothtInc(deltaYaw, self.neckRotTime, self.neckRotMaxSpeed, dt)
            self.neckYaw += EyeHeadGazeLogic.SmoothtInc(-self.neckYaw, self.neckRotTime / self.neckRestTendencyStrenght, self.neckRotMaxSpeed * self.neckRestTendencyStrenght, dt)

            if self.neckYaw > EyeHeadGazeLogic.NECK_MAX_YAW:
                self.neckYaw = EyeHeadGazeLogic.NECK_MAX_YAW
            elif self.neckYaw < EyeHeadGazeLogic.NECK_MIN_YAW:
                self.neckYaw = EyeHeadGazeLogic.NECK_MIN_YAW

            neckPitchRad = self.neckPitch * math.pi / 180
            neckYawRad = self.neckYaw * math.pi / 180

            #
            # Update the global neck transform (G), composing its initial rotation (I) with the rotation (R) composing neck angles yaw and pitch (R = Ry * Rp).
            # G = R * I

            # First compute the absolute-frame delta (D) for the initial pose to the current (C)
            # C = D * I
            # D = C * Iinv
            neckAbsDelta = inOutNeckRotation * self.aPoseNeckRotation.inverted()

            # The rotation of the neck around the global vertical must be calculated
            # be first bringing the neck to its original position and then back to the current pose.
            # Ry = D * Y * Dinv
            yawQuat = Quaternion(EyeHeadGazeLogic.HORIZ_PLANE_NORMAL, -neckYawRad)  # type: Quaternion
            yawGlobal = neckAbsDelta * yawQuat * neckAbsDelta.inverted()  # type: Quaternion

            # Rp = D * P * Dinv
            pitchQuat = Quaternion(EyeHeadGazeLogic.RIGHT, -neckPitchRad)  # type: Quaternion
            pitchGlobal = neckAbsDelta * pitchQuat * neckAbsDelta.inverted()  # type: Quaternion

            # G = Rp * Ry * I
            newNeckRot = pitchGlobal * yawGlobal * self.aPoseNeckRotation

            inOutNeckRotation[:] = newNeckRot[:]

        #
        # Calculate the proportion from the pitch angle to BlendShape percentage
        # weightEyesPitch : 50% = pitch : 45deg
        weightEyesPitch = 0.5 * self.eyesPitch / 45.0

        # Calculate the proportion from the yaw angle to the BlendShape percentage
        # weightEyesYaw : 50% = yaw : 52deg
        weightEyesYaw = 0.5 * self.eyesYaw / 52.0

        # Blendshapes are registered int the DefaultExecutionOrder: EYES_RIGHT, EYES_LEFT, EYES_DOWN, EYES_UP
        if weightEyesPitch >= 0.0:
            outBlendShapeWeights[3] = weightEyesPitch
            outBlendShapeWeights[2] = 0.0
        else:
            outBlendShapeWeights[2] = -weightEyesPitch
            outBlendShapeWeights[3] = 0.0

        if weightEyesYaw >= 0.0:
            outBlendShapeWeights[0] = weightEyesYaw
            outBlendShapeWeights[1] = 0.0
        else:
            outBlendShapeWeights[1] = -weightEyesYaw
            outBlendShapeWeights[0] = 0.0

        self.perf_meter.iterEnd()
