from perf_utils import PerfMeter

#
# Tools to parse the MaryTTS duration files and synthesise them into visemes

RAMP_DOWN_PROPORTION = 1.5
DEFAULT_RAMP_UP_SPEED = 10.0  # How many units per second I change the blend shapes
DEFAULT_RAMP_DOWN_SPEED = DEFAULT_RAMP_UP_SPEED * RAMP_DOWN_PROPORTION

# The viseme will be brought to full weight in this fraction of its duration.
# Range: [0.0,1.0]
RAMP_UP_DURATION_PROPORTION = 0.5

MIN_RAMP_SPEED = 8.0
MAX_RAMP_SPEED = 12.0

ANTICIPATION_SECS = 0.3

# Assumed duration when no following phoneme is present (useful for the last of the sequence.
DEFAULT_VISEME_DURATION_SECS = 0.2


class MaryTTSException(Exception):
    pass


class MaryTTSBlendSequencer:

    # MBLab 1.5
    # VISEMES = ['Etc',
    #            'G',
    #            'SH',
    #            'OO',
    #            'FV',
    #            'TH',
    #            'R',
    #            'L',
    #            'EH',
    #            'O',
    #            'MBP',
    #            'AH',
    #            'S',
    #            'EE']

    # MBLab 1.6
    VISEMES = [
        "phoneme_a_01",
        "phoneme_a_02",
        "phoneme_b_01",
        "phoneme_b_02",
        "phoneme_c_01",
        "phoneme_c_02",
        "phoneme_d_01",
        "phoneme_e_01",
        "phoneme_f_01",
        "phoneme_g_01",
        "phoneme_i_01",
        "phoneme_i_02",
        "phoneme_k_01",
        "phoneme_l_01",
        "phoneme_m_01",
        "phoneme_n_01",
        "phoneme_o_01",
        "phoneme_o_02",
        "phoneme_p_01",
        "phoneme_q_01",
        "phoneme_r_01",
        "phoneme_r_02",
        "phoneme_s_01",
        "phoneme_t_01",
        "phoneme_u_01",
        "phoneme_w_01",
        "phoneme_z_01"
    ]

    # MBLab 1.5
    # PHONEMES_MAP = {'gi': 'G',
    #                 'ge': 'G',
    #                 'ji': 'G',
    #                 'c': 'G',
    #                 'il': 'Etc',  # 'L',
    #                 'el': 'Etc',  # 'L',
    #                 # 'n': 'L',
    #                 'di': 'TH',
    #                 'eh': 'EH',
    #                 # 'i' : 'EH',
    #                 # 'e' : 'EE',
    #                 'ie': 'EE',
    #                 'ee': 'EE',
    #                 'sh': 'SH',
    #                 'sch': 'SH',
    #                 's': 'S',
    #                 'fv': 'FV',
    #                 'fw': 'FV',
    #                 'er': 'R',
    #                 'o': 'O',
    #                 'O': 'O',
    #                 'ov': 'O',
    #                 'oo': 'OO',
    #                 'oh': 'OO',
    #                 'mb': 'MBP',
    #                 'p': 'MBP',
    #                 'ah': 'AH',
    #                 'a': 'AH',
    #
    #                 '{': 'EH',
    #                 '@U': 'O',  # TODO -- split in O-OO, like in 'goes'
    #                 'A': 'AH',
    #                 'AI': 'AH',  # TODO -- split to AH-EE
    #                 'aU': 'AH',  # TODO -- plit into AH-OO
    #                 'b': 'MBP',
    #                 # 'd': '',  #
    #                 'D': 'TH',  # Like in 'the'
    #                 # 'dZ'
    #
    #                 'E': 'EH',
    #                 'EI': 'EE',
    #
    #                 # 'h':
    #                 'i': 'EE',
    #                 'I': 'EE',
    #                 # 'k': ''
    #                 'm': 'MBP',
    #                 'f': 'FV',
    #                 # 'g': '', # This is guttural GH
    #                 'l': 'Etc',  # TODO - This will map to 'L' when we implement tongue and collar muscles
    #                 # 'n': '', #
    #                 # 'N': '', #
    #                 'r': 'R',
    #                 'r=': 'R',
    #                 'S': 'SH',
    #                 # 'T': 'TH',  # TODO - this will use tongue, when available.
    #                 # 't': '', # not tongue between the teeth
    #                 'u': 'OO',
    #                 'v': 'FV',
    #                 'V': 'O',
    #                 'w': 'OO',
    #                 'z': 'S',
    #
    #                 '_': None  # "End of sequence"
    #                 }

    # MBLab 1.6
    PHONEMES_MAP = {'gi': 'phoneme_g_01',
                    'ge': 'phoneme_g_01',
                    'ji': 'phoneme_g_01',
                    'c': 'phoneme_g_01',
                    'il': 'phoneme_l_01',  # 'L',
                    'el': 'phoneme_l_01',  # 'L',
                    # 'n': 'L',
                    'di': 'phoneme_d_01',
                    'eh': 'phoneme_e_01',
                    # 'i' : 'EH',
                    # 'e' : 'EE',
                    'ie': 'phoneme_i_01',
                    'ee': 'phoneme_i_01',
                    'sh': 'phoneme_g_01',
                    'sch': 'phoneme_c_02',
                    's': 'phoneme_e_01',
                    'fv': 'phoneme_f_01',
                    'fw': 'phoneme_f_01',
                    'er': 'phoneme_r_01',
                    'o': 'phoneme_o_01',
                    'O': 'phoneme_o_01',
                    'ov': 'phoneme_o_01',  # 'O',
                    'oo': 'phoneme_u_01',  # 'OO',
                    'oh': 'phoneme_u_01',  # 'OO',
                    'mb': 'phoneme_m_01',  # 'MBP',
                    'p': 'phoneme_p_01',  # 'MBP',
                    'ah': 'phoneme_a_01',  # 'AH',
                    'a': 'phoneme_a_01',  # 'AH',

                    '{': 'phoneme_e_01', # 'EH',
                    '@U': 'phoneme_o_01', #'O',  # TODO -- split in O-OO, like in 'goes'
                    'A': 'phoneme_a_01', #'AH',
                    'AI': 'phoneme_a_01', #'AH',  # TODO -- split to AH-EE
                    'aU': 'phoneme_a_01', #'AH',  # TODO -- split into AH-OO
                    'b': 'phoneme_b_02', #'MBP',
                    'd': 'phoneme_d_01', #'',
                    'D': 'phoneme_d_01', #'TH',   # Like in 'the'
                    'dZ': 'phoneme_d_01',

                    'E': 'phoneme_e_01', #'EH',
                    'EI': 'phoneme_i_01', #'EE',

                    # 'h'=>
                    'i': 'phoneme_i_01', #'EE',
                    'I': 'phoneme_i_01', #'EE',
                    # 'k': ''
                    'm': 'phoneme_m_01', #'MBP',
                    'f': 'phoneme_f_01', #'FV',
                    # 'g': '', # This is guttural GH
                    'l': 'phoneme_l_01', #'Etc',  # TODO - This will map to 'L' when we implement tongue and collar muscles
                    # 'n': '', #
                    # 'N': '', #
                    'r': 'phoneme_r_01', #'R',
                    'r=': 'phoneme_r_01', #'R',
                    'S': 'phoneme_c_01', #'SH',
                    # 'T': 'TH',  # TODO - this will use tongue, when available.
                    # 't': '', # not tongue between the teeth
                    'u': 'phoneme_u_01', #'OO',
                    'v': 'phoneme_b_01', #'FV',
                    'V': 'phoneme_o_01', #'O',
                    'w': 'phoneme_w_01', #'OO',
                    'z': 'phoneme_z_01', #'S',

                    '_': None  # "End of sequence"
                    }


    @classmethod
    def get_viseme_count(cls):
        """Returns the number of Visemes which this module expects to handle."""
        return len(MaryTTSBlendSequencer.VISEMES)

    def __init__(self):
        """This class can be used to parse a MaryTTS realized durations file and also to generate in real-time the
        weights of the blend shapes."""

        self.perf_meter = PerfMeter("MaryTTSPythonManual", updateInterval=10.0, writeToFile=True)

        self.realized_times = None
        self.realized_visemes = None

        self.last_time = None
        self.speak_start_time = None
        self.position_tts = 0

        self.ready_to_speak = False  # This will be set true after loading a sentence or resetting the timer.
        self.active_viseme = None

        self.ramp_up_speed = DEFAULT_RAMP_UP_SPEED
        self.ramp_down_speed = DEFAULT_RAMP_DOWN_SPEED

        # Consistency check: all the table values MUST be in the MOUTH_SHAPES set.
        for ph in MaryTTSBlendSequencer.PHONEMES_MAP:
            ms = MaryTTSBlendSequencer.PHONEMES_MAP[ph]
            if ms is not None and ms not in MaryTTSBlendSequencer.VISEMES:
                raise Exception("For phoneme {}, target mouth shape {} not listed.".format(ph, ms))

    def stop_sequencer(self):
        self.ready_to_speak = False

    def parse_realized_durations(self, realized_durations: str):
        """Parse a MaryTTS realized durations file and fills object vectors for later sequencing.

        :param realized_durations: The MaryTTS durations file as single multi-line string
        :return: None
        """

        self.realized_times = []
        self.realized_visemes = []

        parsed_durations = []
        parsed_phonemes = []

        # storing phonemes and durations from marytts file to arrays
        for line in realized_durations.split('\n'):
            if len(line) == 0:  # Skip empty lines
                continue
            if line[0] == '#':  # Skip comments
                continue
            split_line = line.split(None)
            if len(split_line) < 3:
                raise MaryTTSException("not enough data in line: '"+line+"'")

            parsed_durations.append(split_line[0])  # duration of the phoneme
            parsed_phonemes.append(split_line[2])   # phoneme

        assert len(parsed_durations) == len(parsed_phonemes)
        # print("PARSED ("+str(len(parsed_durations))+"):")
        # print("Durations:")
        # print(parsed_durations)
        # print("Phonemes:")
        # print(parsed_phonemes)

        # Mapping phonemes
        for i, p in enumerate(parsed_phonemes):
            # p = p.lower()
            dur = float(parsed_durations[i])

            if p in MaryTTSBlendSequencer.PHONEMES_MAP:
                shape = MaryTTSBlendSequencer.PHONEMES_MAP[p]
            else:
                shape = 'Etc'

            # print("Mapping '{}' --> '{}' at {}".format(p, shape, dur))

            self.realized_times.append(dur)
            self.realized_visemes.append(shape)

        self.ready_to_speak = True

        assert len(self.realized_visemes) == len(self.realized_times)
        # print("MAPPED (" + str(len(self.realized_visemes)) + "):")
        # print(self.realized_visemes)
        # print(self.realized_times)

    def reset_timers(self):
        """Reset the internal timers. This must b called if you want to synthetise the same sentence again
         without re-parsing the durations."""

        self.last_time = None
        self.speak_start_time = None
        self.position_tts = 0
        self.ready_to_speak = True
        self.active_viseme = None

    def update(self, now: float, viseme_weights: list):
        """The high-frequency routine to call 30+ times per second.

        :param now: the current time, in seconds.
        :param viseme_weights: An array of current weights of the viseme blend shapes.
         Values will be updated according to the synthesis time.
         The size of this vector must be the same of the module static vector of visemes.
         Values will be clamped in the range [0,1].
        :return: None
        """

        self.perf_meter.iterStart()

        if :
            return

        if len(viseme_weights) != len(MaryTTSBlendSequencer.VISEMES):
            raise MaryTTSException("Viseme weight vector has wrong size."
                                   " Expected {}, found {}".format(len(MaryTTSBlendSequencer.VISEMES), len(viseme_weights)))

        # now = time.time()  # current time

        if self.last_time is None:
            self.last_time = now
            self.speak_start_time = now
            self.position_tts = 0
            return

        elapsed_time = now - self.speak_start_time  # time elapsed
        delta_time = now - self.last_time

        self.last_time = now

        elapsed_time += ANTICIPATION_SECS

        #
        # CHECK WHICH PHONEME WE HAVE TO USE
        phonemes_count = len(self.realized_visemes)  # how many actions

        if self.position_tts < phonemes_count:
            time_to_wait = self.realized_times[self.position_tts]
            # time_to_wait -= ANTICIPATION_SECS
            if elapsed_time >= time_to_wait:
                # Current action
                self.active_viseme = self.realized_visemes[self.position_tts]
                # And advance to the next phoneme
                self.position_tts += 1

                # Look ahead time, and adjust ramp speeds
                viseme_duration = DEFAULT_VISEME_DURATION_SECS
                if self.position_tts < phonemes_count:
                    next_time_to_wait = self.realized_times[self.position_tts]
                    # print("ww={}, next {}".format(time_to_wait, next_time_to_wait))
                    viseme_duration = next_time_to_wait - time_to_wait

                # How long has to last the ramp duration
                ramp_up_duration = viseme_duration * RAMP_UP_DURATION_PROPORTION
                # Ramp up speed. (Trivial formula, since the range is 0 to 1
                self.ramp_up_speed = 1 / ramp_up_duration
                self.ramp_up_speed = sorted((MIN_RAMP_SPEED, self.ramp_up_speed, MAX_RAMP_SPEED))[1]  # clamp :-)
                self.ramp_down_speed = self.ramp_up_speed * RAMP_DOWN_PROPORTION

                # print("Switching to viseme {}, duration {}, ramp_up_dur={} -> speed={}".format(self.active_viseme, viseme_duration, ramp_up_duration, self.ramp_up_speed))

        else:
            self.active_viseme = None
            last_time_to_wait = 0
            if len(self.realized_times) > 0:
                last_time_to_wait = self.realized_times[-1]
            if elapsed_time > (last_time_to_wait + 1.0):
                self.ready_to_speak = False

        #
        # Debug - print nicely formatted table with blend weights
        tab_template = "{:3s} "
        tab_template += "{:.2f}  " * len(viseme_weights)
        # print(tab_template.format(self.active_viseme if self.active_viseme is not None else "--", *viseme_weights))

        #
        # UPDATE THE FACE
        for i, viseme in enumerate(MaryTTSBlendSequencer.VISEMES):
            current_weight = viseme_weights[i]
            # ch = obj.channels['mouth-' + phoneme]
            if viseme == self.active_viseme:  # This has to ramp up to 1.0
                inc = self.ramp_up_speed * delta_time
                # print("delta={}, inc={}, y={}, ly={}".format(delta_time, inc, y, ch.location.y))
                # ch.location.y += inc  # This doesn't work. I have to extract the y and set the triplet altogether.
                current_weight += inc
                if current_weight > 1.0:
                    current_weight = 1.0
            else:  # This ha to ramp down to 0.0
                dec = - self.ramp_down_speed * delta_time
                # print("delta={}, inc={}, y={}, ly={}".format(delta_time, dec, y, ch.location.y))
                current_weight += dec
                if current_weight < 0.0:
                    current_weight = 0.0

            # update the weight
            viseme_weights[i] = current_weight

        self.perf_meter.iterEnd()
